use std::{net::TcpStream, io::{Read, Write}};
use serde::Serialize;

/// version of protocol
const JSONRPC_VERSION: &str = "2.0";
const SERVICE_ADDRESS: &str = "127.0.0.1:43424";

#[derive(Clone, Debug, Serialize)]
pub struct Params<T: Serialize>(pub T);

#[derive(Clone, Debug, Serialize)]
pub struct Request<T: Serialize> {
    jsonrpc: & 'static str,
    pub method: String,
    pub params: Params<T>,
    pub id: i32,
}

#[derive(Debug)]
pub struct RequestError {
    pub error: String,
}

impl<T: Serialize> Request<T> {

    pub fn new(method: &str, params: Params<T>, id: i32) -> Request<T> {
        Request {
            jsonrpc: JSONRPC_VERSION,
            method: method.to_owned(),
            params,
            id,
        }
    }

    pub fn to_string(&self) -> String
    {
        serde_json::to_string(&self).unwrap()
    }

    pub fn send(&self) -> Result<serde_json::Value, RequestError>
    {
        let mut client = TcpStream::connect(&SERVICE_ADDRESS).unwrap();

        let json = serde_json::to_string_pretty(&self).unwrap();

        // send request to the API
        match client.write_all(json.as_bytes()) {
            Ok(_) => (),
            Err(error) => {
                return Err(RequestError{error: error.to_string()})
            }
        }

        let mut buffer = [0u8; 4 * 1024];

        // read the response
        let buffer_size = match client.read(&mut buffer) {
            Ok(size) => size,
            Err(error) => {
                return Err(RequestError{error: error.to_string()})
            }
        };

        let mut result: serde_json::Value = match serde_json::from_str(
            String::from_utf8_lossy(
                &buffer[..buffer_size]
            ).trim())
        {
            Ok(result) => result,
            Err(error) =>
            {
                return Err(RequestError { error: error.to_string() })
            }
        };

        if result["error"].is_object()
        {
            let error = &result["error"]["message"];

            return Err(RequestError { error: format!("API error : {}", error) });
        };

        Ok(result["result"].take())
    }

    // Send and expect u64 result
    pub fn as_u64(&self) -> Option<u64>
    {
        match self.send()
        {
            Ok(result) => result.as_u64(),
            Err(_) => None
        }
    }

    // Send and expect bool result
    pub fn as_bool(&self) -> Option<bool>
    {
        match self.send()
        {
            Ok(result) => result.as_bool(),
            Err(_) => None
        }
    }

    // Send and expect string result
    pub fn as_string(&self) -> Option<String>
    {
        match self.send()
        {
            Ok(result) if result.is_string() => Some(result.as_str().unwrap().to_owned()),
            Ok(_result) => return None, // Not the expected type
            Err(_) => return None,
        }
    }
}